#!/usr/bin/env python3
# coding=utf-8

import argparse
import sys

from romt import common
from romt import error
import romt.crate
import romt.rustup
import romt.serve
import romt.toolchain

__version__ = "0.3.1"

project_name = "romt"

description = """\
Rust Offline Mirror Tool
"""

epilog = """\
See ``{project_name} --readme`` for more details.
""".format(
    project_name=project_name
)


def readme_from_pkg_resources() -> str:
    # This method works when the package is properly installed via "pip".
    import pkg_resources
    import email
    import textwrap

    try:
        dist = pkg_resources.get_distribution(project_name)
        meta = dist.get_metadata(dist.PKG_INFO)
    except (pkg_resources.DistributionNotFound, FileNotFoundError):
        return ""
    msg = email.message_from_string(meta)
    desc = msg.get("Description", "").strip()
    if not desc and not msg.is_multipart():
        desc = msg.get_payload().strip()
    if "\n" in desc:
        first, rest = desc.split("\n", 1)
        desc = "\n".join([first, textwrap.dedent(rest)])
    return desc


def readme_from_file() -> str:
    # This method works with PyInstaller.
    import pkgutil

    text = ""
    try:
        readme = pkgutil.get_data(project_name, "README.rst")
        if readme is not None:
            text = readme.decode("utf-8")
    except FileNotFoundError:
        text = ""
    return text


def readme() -> None:
    desc = readme_from_pkg_resources()
    if not desc:
        desc = readme_from_file()
    if not desc:
        desc = "README.rst is not available."
    common.iprint(desc)


def add_common_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__
    )

    parser.add_argument(
        "-v", "--verbose", action="count", default=0, help="verbose output",
    )

    parser.add_argument(
        "-q", "--quiet", action="count", default=0, help="quiet output",
    )

    parser.add_argument(
        "--readme", action="store_true", help="display README.rst",
    )

    parser.add_argument(
        "--num-jobs",
        "-j",
        type=int,
        default=4,
        action="store",
        help="number of simultaneous download jobs (default %(default)s)",
    )


def make_parser() -> argparse.ArgumentParser:
    common_parser = argparse.ArgumentParser(add_help=False)
    add_common_arguments(common_parser)

    parser = argparse.ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[common_parser],
    )

    subparsers = parser.add_subparsers(dest="subparser_name", help="OPERATION")

    crate_parser = subparsers.add_parser(
        "crate",
        help=romt.crate.description,
        description=romt.crate.description,
        epilog=romt.crate.epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[common_parser],
    )

    romt.crate.add_arguments(crate_parser)

    rustup_parser = subparsers.add_parser(
        "rustup",
        help=romt.rustup.description,
        description=romt.rustup.description,
        epilog=romt.rustup.epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[common_parser],
    )

    romt.rustup.add_arguments(rustup_parser)

    toolchain_parser = subparsers.add_parser(
        "toolchain",
        help=romt.toolchain.description,
        description=romt.toolchain.description,
        epilog=romt.toolchain.epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[common_parser],
    )

    romt.toolchain.add_arguments(toolchain_parser)

    serve_parser = subparsers.add_parser(
        "serve",
        help=romt.serve.description,
        description=romt.serve.description,
        epilog=romt.serve.epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[common_parser],
    )

    romt.serve.add_arguments(serve_parser)

    return parser


def main() -> None:
    parser = make_parser()
    args = parser.parse_args()
    common.set_max_verbosity(common.VERBOSITY_INFO + args.verbose - args.quiet)

    try:
        cmd = args.subparser_name
        if args.readme:
            readme()
        elif cmd is None:
            raise error.UsageError("missing OPERATION (try --help)")
        elif cmd == "crate":
            romt.crate.Main(args).run()
        elif cmd == "rustup":
            romt.rustup.Main(args).run()
        elif cmd == "toolchain":
            romt.toolchain.Main(args).run()
        elif cmd == "serve":
            romt.serve.Main(args).run()

    except error.Error as e:
        common.eprint(e)
        sys.exit(1)

    except KeyboardInterrupt:
        common.eprint("Keyboard interrupt")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding=utf-8

import argparse
from pathlib import Path
from typing import List, Optional

from romt import error
import romt.download


def verify_commands(commands: List[str], valid_commands: List[str]) -> None:
    for command in commands:
        if command not in valid_commands:
            raise error.UsageError("invalid COMMAND {}".format(repr(command)))


def add_downloader_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--assume-ok",
        action="store_true",
        default=False,
        help="assume already-downloaded files are OK (skip hash check)",
    )


class BaseMain:
    def __init__(self, args: argparse.Namespace) -> None:
        self.args = args
        self._downloader = None  # type: Optional[romt.download.Downloader]

    @property
    def downloader(self) -> romt.download.Downloader:
        if self._downloader is None:
            num_jobs = max(self.args.num_jobs, 1)
            self._downloader = romt.download.Downloader(num_jobs=num_jobs)
        return self._downloader

    def get_archive_path(self) -> Path:
        if not self.args.archive:
            raise error.UsageError("missing archive name")
        return Path(self.args.archive)

    def _run(self) -> None:
        # Override in derived classes.
        pass

    def run(self) -> None:
        try:
            self._run()
        finally:
            if self._downloader is not None:
                self._downloader.close()
                self._downloader = None



#!/usr/bin/env python3
# coding=utf-8

import argparse
from pathlib import Path
import re
import shutil
from typing import (
    Iterable,
    List,
    Set,
    Tuple,
)

import trio

from romt import base
from romt import common
from romt import dist
from romt import error
from romt import integrity
from romt import signature
from romt.manifest import Manifest, Package

description = """\
Mirror and manage toolchain artifacts.
"""

epilog = """\

SPEC is one of:
  <channel>
  <channel>-YYYY-MM-DD
  <channel>-latest
  <channel>-*
  *-YYYY-MM-DD
  *-latest
  *-*
  YYYY-MM-DD
  latest
  *
Where:
- <channel> is one of: nightly, beta, stable, X.Y.Z
- Single ``*`` represents date, not channel; equivalent to ``*-*``.
- Multiple SPEC options may be given.
- Each SPEC option will be split at commas and whitespace.

TARGET is standard 3- or 4-tuple; common examples:
- x86_64-unknown-linux-gnu (alias ``linux``)
- x86_64-pc-windows-msvc (alias ``windows``)
- x86_64-apple-darwin (alias ``darwin``)

COMMAND values:
  Typical:
    download        download artifacts matching SPEC and TARGET to DEST
    pack            pack DEST artifacts matching SPEC and TARGET into ARCHIVE
    unpack          unpack ARCHIVE into DEST, setting SPEC and TARGET
  Less common:
    fetch-manifest  download manifest matching SPEC and TARGET to DEST
    verify          verify DEST artifacts matching SPEC and TARGET
    list            print DEST artifacts and targets matching SPEC
    all-targets     print all known targets mentioned in SPEC
    fixup           publish manifest variants for SPEC, updating undated

When multiple COMMANDs are given, they share all option values.

For complete details, try ``romt --readme`` to view README.rst.
"""


def parse_spec(spec: str) -> Tuple[str, str]:
    """parse spec into (date, channel).

    Forms with channel:
      <channel>
      <channel>-YYYY-MM-DD
      <channel>-latest
      <channel>-*
      *-YYYY-MM-DD
      *-latest
      *-*
    Forms with only date:
      YYYY-MM-DD
      latest
      *

    Where:
    - <channel> is one of: nightly, beta, stable, X.Y.Z
    - Single ``*`` represents date, not channel; equivalent to ``*-*``.

    """

    # Single "*" is treated as wildcard for date, not channel.
    if spec == "*":
        return "*", "*"

    channel_rex = r"""
        (?P<channel>
            nightly | beta | stable | \* | (?: \d+\.\d+\.\d+ )
        )
        """

    date_rex = r"""
        (?P<date>
            \d\d\d\d-\d\d-\d\d | latest | \*
        )
        """

    m = re.match(
        r"{} (?: - {})? $".format(channel_rex, date_rex), spec, re.VERBOSE
    )
    if m:
        channel = m.group("channel")
        date = m.group("date") or ""
        return date, channel

    m = re.match(r"{} $".format(date_rex), spec, re.VERBOSE)
    if m:
        date = m.group("date")
        return date, "*"

    raise error.UsageError("invalid SPEC {}".format(repr(spec)))


def channel_rel_path(date: str, channel: str) -> str:
    channel_name = "channel-rust-{}.toml".format(channel)
    if date:
        rel_path = "{}/{}".format(date, channel_name)
    else:
        rel_path = "{}".format(channel_name)
    return rel_path


def add_arguments(parser: argparse.ArgumentParser) -> None:
    base.add_downloader_arguments(parser)

    parser.add_argument(
        "-s",
        "--select",
        action="append",
        dest="specs",
        metavar="SPEC",
        default=[],
        help="""one or more SPEC values for toolchain selection""",
    )

    parser.add_argument(
        "-t",
        "--target",
        action="append",
        dest="targets",
        default=[],
        help="""target to download (default varies by COMMAND)""",
    )

    parser.add_argument(
        "--dest",
        action="store",
        default="dist",
        help="""local download directory (default: %(default)s)""",
    )

    parser.add_argument(
        "--url",
        action="store",
        default="https://static.rust-lang.org/dist",
        help="""base URL of dist (default: %(default)s)""",
    )

    parser.add_argument(
        "--archive",
        action="store",
        metavar="ARCHIVE",
        default="toolchain.tar.gz",
        help="use archive ARCHIVE for pack/unpack (default: %(default)s)",
    )

    parser.add_argument(
        "--warn-signature",
        action="store_true",
        default=False,
        help="warn (instead of fail) on signature verification failure",
    )

    parser.add_argument(
        "--no-signature",
        action="store_true",
        default=False,
        help="disable all uses of signature files (*.asc); mainly for testing",
    )

    parser.add_argument(
        "commands",
        nargs="*",
        metavar="COMMAND",
        help="""commands to execute in the order given""",
    )


class Main(dist.DistMain):
    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.downloader.set_warn_signature(args.warn_signature)
        self._with_sig = not args.no_signature

    def manifest_url_path(self, date: str, channel: str) -> Tuple[str, Path]:
        rel_path = channel_rel_path(date, channel)
        manifest_url = self.url_from_rel_path(rel_path)
        manifest_path = self.dest_path_from_rel_path(rel_path)
        return manifest_url, manifest_path

    def manifest_path(self, date: str, channel: str) -> Path:
        _, man_path = self.manifest_url_path(date, channel)
        return man_path

    def get_manifest(
        self, date: str, channel: str, *, download: bool
    ) -> Manifest:
        man_url, man_path = self.manifest_url_path(date, channel)
        if download:
            if date:
                # Dated manifests may always be cached.
                cached = True
            else:
                # Undated manifests should always be re-downloaded.  They
                # might be newer than what's on disk (either because of the
                # passage of time or because a fixup operation might overwrite
                # the undated manifest with old information before a download).
                cached = False
            self.downloader.download_verify(
                man_url, man_path, cached=cached, with_sig=self._with_sig
            )
        else:
            self.downloader.verify(man_path, with_sig=self._with_sig)
        return Manifest.from_toml_path(man_path)

    def select_manifest(
        self, spec: str, *, download: bool, canonical: bool = False
    ) -> Manifest:
        date, channel = parse_spec(spec)
        manifest = self.get_manifest(date, channel, download=download)
        if canonical and (
            manifest.date != date or manifest.channel != channel
        ):
            manifest = self.get_manifest(
                manifest.date, manifest.channel, download=download
            )
        return manifest

    def channels_in_dest_date(self, date: str) -> List[str]:
        date_path = self.dest_path / date
        prefix = "channel-rust-"
        suffix = ".toml"
        channels = [
            p.name[len(prefix) : -len(suffix)]
            for p in date_path.glob(prefix + "*" + suffix)
        ]
        return channels

    def adjust_download_specs(self, specs: List[str]) -> List[str]:
        # For downloads, require explicit date and channel.
        for spec in specs:
            date, channel = parse_spec(spec)
            if "*" in (date, channel) or date == "latest":
                raise error.UsageError("invalid wild SPEC: {}".format(spec))
        return dist.require_specs(specs)

    def expand_wild_spec(self, spec: str) -> List[str]:
        specs = []  # type: List[str]
        date, channel = parse_spec(spec)
        if "*" in (date, channel) or date == "latest":
            if channel == "*":
                channel_patterns = set("nightly beta stable".split())
            else:
                channel_patterns = set([channel])
            if date in ("*", "latest"):
                dates = common.reversed_date_dir_names(self.dest_path)
            else:
                dates = [date]
            for d in dates:
                channels = channel_patterns.intersection(
                    self.channels_in_dest_date(d)
                )
                specs.extend(
                    "{}-{}".format(channel, d) for channel in channels
                )
                if date == "latest" and specs:
                    break
        else:
            specs.append(spec)

        if not specs:
            raise error.UsageError(
                "no matches for wild SPEC {}".format(repr(spec))
            )

        return specs

    def adjust_wild_specs(self, specs: List[str]) -> List[str]:
        # For non-downloads, handle wild specs.
        adjusted_specs = []
        for spec in specs:
            adjusted_specs.extend(self.expand_wild_spec(spec))
        return dist.require_specs(adjusted_specs)

    def downloaded_packages(self, manifest: Manifest) -> List[Package]:
        packages = []
        for package_info in manifest.gen_available_packages():
            dest_path = self.dest_path_from_rel_path(package_info.rel_path)
            if dest_path.is_file():
                packages.append(package_info)
        return packages

    def downloaded_targets(self, manifest: Manifest) -> List[str]:
        targets = set(p.target for p in self.downloaded_packages(manifest))
        targets.discard("*")
        return sorted(targets)

    def adjust_targets(
        self, manifest: Manifest, base_targets: List[str]
    ) -> List[str]:
        possible_targets = set(p.target for p in manifest.gen_packages())
        targets = set()
        for target in base_targets:
            if target == "all":
                targets.update(possible_targets)
            elif target == "*":
                targets.update(self.downloaded_targets(manifest))
            elif target not in possible_targets:
                raise error.UsageError(
                    "target {} not found in manifest".format(repr(target))
                )
            else:
                targets.add(target)
        return sorted(targets)

    def cmd_fetch_manifest(self) -> None:
        for spec in self.adjust_download_specs(self.specs):
            common.iprint("Fetch manifest: {}".format(spec))
            manifest = self.select_manifest(spec, download=True)
            common.iprint("  ident: {}".format(manifest.ident))

    async def _download_verify_one(
        self,
        limiter: trio.CapacityLimiter,
        download: bool,
        dest_url: str,
        dest_path: Path,
    ) -> None:
        try:
            if download:
                await self.downloader.adownload_verify(
                    dest_url,
                    dest_path,
                    assume_ok=self.args.assume_ok,
                    with_sig=self._with_sig,
                )
            else:
                self.downloader.verify(dest_path, with_sig=self._with_sig)
        finally:
            limiter.release_on_behalf_of(dest_path)

    async def _download_verify_packages(
        self, download: bool, packages: Iterable[Package]
    ) -> None:
        async with trio.open_nursery() as nursery:
            limiter = self.downloader.new_limiter()
            for package in packages:
                rel_path = package.rel_path
                dest_path = self.dest_path_from_rel_path(rel_path)
                dest_url = self.url_from_rel_path(rel_path)
                await limiter.acquire_on_behalf_of(dest_path)
                nursery.start_soon(
                    self._download_verify_one,
                    limiter,
                    download,
                    dest_url,
                    dest_path,
                )

    def _download_verify(
        self, download: bool, specs: List[str], base_targets: List[str]
    ) -> None:
        for spec in specs:
            common.iprint(
                "{}: {}".format("Download" if download else "Verify", spec)
            )
            manifest = self.select_manifest(
                spec, download=download, canonical=True
            )
            common.iprint("  ident: {}".format(manifest.ident))

            targets = self.adjust_targets(manifest, base_targets)
            packages = list(manifest.gen_available_packages(targets=targets))
            common.iprint(
                "  packages: {}, targets: {}".format(
                    len(packages), len(targets)
                )
            )
            for t in targets:
                common.vvprint("  target: {}".format(t))

            self.downloader.run_job(
                self._download_verify_packages, download, packages
            )

    def cmd_download(self) -> None:
        specs = self.adjust_download_specs(self.specs)
        base_targets = dist.require_targets(self.targets)
        self._download_verify(
            download=True, specs=specs, base_targets=base_targets
        )

    def cmd_verify(self) -> None:
        specs = self.adjust_wild_specs(self.specs)
        base_targets = dist.require_targets(self.targets, default="*")
        self._download_verify(
            download=False, specs=specs, base_targets=base_targets
        )

    def cmd_list(self) -> None:
        max_verbosity = common.get_max_verbosity()
        show_details = max_verbosity >= common.VERBOSITY_INFO
        for spec in self.adjust_wild_specs(self.specs):
            common.vprint("List: {}".format(spec))
            manifest = self.select_manifest(spec, download=False)
            if show_details:
                available_packages = manifest.available_packages()
                available_targets = manifest.available_targets()
                packages = self.downloaded_packages(manifest)
                targets = self.downloaded_targets(manifest)

                target_out = "targets[{}/{}]".format(
                    len(targets), len(available_targets),
                )
                package_out = "packages[{}/{}]".format(
                    len(packages), len(available_packages),
                )
                # Example output:
                #   stable-2020-01-30(1.41.0)    \
                #     targets[84/84], packages[272/326]
                common.iprint(
                    "{:28} {:16} {:18}".format(
                        manifest.ident, target_out, package_out
                    )
                )
                for target in targets:
                    common.iprint("  {}".format(target))
            else:
                common.eprint(manifest.ident)

    def cmd_all_targets(self) -> None:
        for spec in self.adjust_wild_specs(self.specs):
            common.iprint("All targets: {}".format(spec))
            manifest = self.select_manifest(spec, download=False)
            common.iprint("  ident: {}".format(manifest.ident))
            for target in manifest.all_targets():
                common.eprint(target)

    def cmd_pack(self) -> None:
        base_targets = dist.require_targets(self.targets, default="*")
        archive_path = self.get_archive_path()
        common.iprint("Packing archive: {}".format(archive_path))
        with common.tar_context(archive_path, "w") as tar_f:

            def pack_path(rel_path: str) -> None:
                dest_path = self.dest_path_from_rel_path(rel_path)
                packed_name = "dist/" + rel_path
                common.vprint("[pack] {}".format(rel_path))
                try:
                    tar_f.add(str(dest_path), packed_name)
                except FileNotFoundError:
                    raise error.MissingFileError(str(dest_path))

            def pack_rel_path(rel_path: str) -> None:
                pack_path(rel_path)
                pack_path(integrity.append_hash_suffix(rel_path))
                if self._with_sig:
                    pack_path(signature.append_sig_suffix(rel_path))

            for spec in self.adjust_wild_specs(self.specs):
                common.iprint("Pack: {}".format(spec))
                manifest = self.select_manifest(
                    spec, download=False, canonical=True
                )
                common.iprint("  ident: {}".format(manifest.ident))

                targets = self.adjust_targets(manifest, base_targets)
                packages = list(
                    manifest.gen_available_packages(targets=targets)
                )
                common.iprint(
                    "  packages: {}, targets: {}".format(
                        len(packages), len(targets)
                    )
                )
                for t in targets:
                    common.vvprint("  target: {}".format(t))

                # Pack channel file.
                pack_rel_path(
                    channel_rel_path(manifest.date, manifest.channel)
                )

                # Pack up package file parts.
                for package in packages:
                    pack_rel_path(package.rel_path)

    def _detect_specs(self, rel_paths: Set[str]) -> List[str]:
        specs = []
        for rel_path in rel_paths:
            m = re.match(
                r"""
                (?P<date>\d\d\d\d-\d\d-\d\d)
                /channel-rust-
                (?P<channel>nightly|stable|beta)
                \.toml$
                """,
                rel_path,
                re.VERBOSE,
            )
            if m:
                date = m.group("date")
                channel = m.group("channel")
                specs.append("{}-{}".format(channel, date))
        return specs

    def _detect_targets(
        self, specs: List[str], rel_paths: Set[str]
    ) -> List[str]:
        targets = set()
        for spec in specs:
            manifest = self.select_manifest(spec, download=False)
            for package in manifest.gen_available_packages():
                if package.target != "*" and package.rel_path in rel_paths:
                    targets.add(package.target)
        return sorted(targets)

    def cmd_unpack(self) -> None:
        archive_path = self.get_archive_path()
        common.iprint("Unpacking archive: {}".format(archive_path))
        dist_prefix = "dist/"
        extracted = set()
        with common.tar_context(archive_path, "r") as tar_f:
            for tar_info in tar_f:
                if tar_info.isdir():
                    continue
                if not tar_info.name.startswith(dist_prefix):
                    raise error.UnexpectedArchiveMemberError(tar_info.name)

                rel_path = tar_info.name[len(dist_prefix) :]
                dest_path = self.dest_path_from_rel_path(rel_path)
                tar_info.name = str(dest_path)
                common.vprint("[unpack] {}".format(rel_path))
                tar_f.extract(tar_info)
                extracted.add(rel_path)

        specs = self._detect_specs(extracted)
        targets = self._detect_targets(specs, extracted)

        common.iprint("Unpacked specs: {}".format(len(specs)))
        for spec in specs:
            common.iprint("  {}".format(spec))

        common.iprint("Unpacked targets: {}".format(len(targets)))
        for target in targets:
            common.iprint("  {}".format(target))

        self.specs = specs
        self.targets = targets

    def _write_manifest(
        self, manifest: Manifest, date: str, channel: str
    ) -> None:
        src_path = self.manifest_path(manifest.date, manifest.channel)
        dst_path = self.manifest_path(date, channel)
        common.iprint("[publish] {}".format(dst_path))
        shutil.copyfile(str(src_path), str(dst_path))
        src_hash_path = integrity.path_append_hash_suffix(src_path)
        dst_hash_path = integrity.path_append_hash_suffix(dst_path)
        shutil.copyfile(str(src_hash_path), str(dst_hash_path))
        src_sig_path = signature.path_append_sig_suffix(src_path)
        dst_sig_path = signature.path_append_sig_suffix(dst_path)
        shutil.copyfile(str(src_sig_path), str(dst_sig_path))

    def _write_manifest_variations(self, manifest: Manifest) -> None:
        date = manifest.date
        channel = manifest.channel

        # Write top-level manifest, unless a newer one already exists, e.g.:
        #   dist/channel-rust-nightly.toml
        top_path = self.manifest_path("", channel)
        if top_path.is_file():
            top_man = self.get_manifest("", channel, download=False)
            write_top = manifest.date >= top_man.date
        else:
            write_top = True
        if write_top:
            self._write_manifest(manifest, "", channel)

        if channel == "stable":
            # Write version-numbered manifests, e.g.:
            #   dist/yyyy-mm-dd/channel-rust-x.y.z.toml
            #   dist/channel-rust-x.y.z.toml
            self._write_manifest(manifest, date, manifest.version)
            self._write_manifest(manifest, "", manifest.version)

    def cmd_fixup(self) -> None:
        for spec in self.adjust_wild_specs(self.specs):
            common.iprint("Fixup: {}".format(spec))
            manifest = self.select_manifest(
                spec, download=False, canonical=True
            )
            common.vprint("  ident: {}".format(manifest.ident))
            self._write_manifest_variations(manifest)

    def _run(self) -> None:
        valid_commands = [
            "fetch-manifest",
            "download",
            "verify",
            "list",
            "all-targets",
            "pack",
            "unpack",
            "fixup",
        ]
        commands = self.args.commands
        if commands:
            base.verify_commands(commands, valid_commands)
        else:
            common.iprint("nothing to do; try a COMMAND")
            return

        for cmd in commands:
            if cmd == "fetch-manifest":
                self.cmd_fetch_manifest()

            elif cmd == "download":
                self.cmd_download()
                self.cmd_fixup()

            elif cmd == "verify":
                self.cmd_verify()

            elif cmd == "list":
                self.cmd_list()

            elif cmd == "all-targets":
                self.cmd_all_targets()

            elif cmd == "pack":
                self.cmd_pack()

            elif cmd == "unpack":
                self.cmd_unpack()
                self.cmd_verify()
                self.cmd_fixup()

            elif cmd == "fixup":
                self.cmd_fixup()



#!/usr/bin/env python3
# coding=utf-8

import argparse
import json
import os
from pathlib import Path
import re
from typing import Generator, List, Optional, Tuple
import urllib.parse

import git
import git.remote
import trio

from romt import base
from romt import common
from romt import error
import romt.download


# Name of index:
INDEX_NAME = "crates.io-index"


# Standard location for index repository:
INDEX_STANDARD_PATH = Path("git") / INDEX_NAME


# Name of index bundle file:
INDEX_BUNDLE_NAME = "origin.bundle"


# Name of index bundle within archive:
INDEX_BUNDLE_PACKED_NAME = "{}/{}".format(
    INDEX_STANDARD_PATH.as_posix(), INDEX_BUNDLE_NAME
)


description = """\
Mirror and manage crates from crates.io.
"""

epilog = """\
COMMAND values:
  Initial setup:
    init            setup INDEX and CRATES_ROOT for download/export scenario
    init-import     setup INDEX and CRATES_ROOT for import scenario
    config          edit INDEX/config.json to point to mirror server
  Routine updates:
    update          alias for ``pull download mark``
    export          alias for ``pull download pack mark``
    import          alias for ``unpack pull verify mark``
  Less common:
    pull            pull {INDEX_NAME} commits from origin into INDEX
    download        download .crate files to CRATES_ROOT across RANGE
    verify          verify .crate files in CRATES_ROOT across RANGE
    pack            pack RANGE of crates and INDEX commits into ARCHIVE
    mark            set branch ``mark`` to match END of RANGE
    unpack          unpack {INDEX_BUNDLE_NAME} and CRATES_ROOT from ARCHIVE
    list            print crate names across RANGE

When multiple COMMANDs are given, they share all option values.

For complete details, try ``romt --readme`` to view README.rst.
""".format(
    INDEX_NAME=INDEX_NAME, INDEX_BUNDLE_NAME=INDEX_BUNDLE_NAME
)


def crate_prefix_from_name(name: str) -> str:
    if len(name) == 1:
        prefix = "1"
    elif len(name) == 2:
        prefix = "2"
    elif len(name) == 3:
        prefix = "3/{}".format(name[0])
    else:
        prefix = "{}/{}".format(name[:2], name[2:4])
    return prefix


def crate_rel_path_from_name_version(name: str, version: str) -> Path:
    prefix = crate_prefix_from_name(name)
    return Path(prefix) / name / "{}-{}.crate".format(name, version)


def blobs_in_commit_range(
    start_commit: git.Commit, end_commit: git.Commit
) -> Generator[Tuple[str, bytes, bytes], None, None]:
    """Generate (blob_path, start_blob, end_blob)."""
    if start_commit:
        for diff in end_commit.diff(start_commit):
            # diff.a_xxx goes with end_commit.
            if diff.a_blob:
                a_blob = diff.a_blob.data_stream.read()
                if diff.b_blob:
                    b_blob = diff.b_blob.data_stream.read()
                else:
                    b_blob = b""
                yield diff.a_path, b_blob, a_blob
    else:
        for item in end_commit.tree.traverse():
            if item.type == "blob":
                yield item.path, b"", item.data_stream.read()


class Crate:
    def __init__(self, name: str, version: str, hash: str):
        self.name = name
        self.version = version
        self.hash = hash

    def __str__(self) -> str:
        return "Crate({}, {}, {})".format(self.name, self.version, self.hash)

    def rel_path(self) -> Path:
        return crate_rel_path_from_name_version(self.name, self.version)


def blob_lines(blob: bytes) -> Generator[str, None, None]:
    for line in blob.decode("utf8").splitlines():
        yield line


def crate_from_text(line: str) -> Crate:
    j = json.loads(line)
    return Crate(j["name"], j["vers"], j["cksum"])


def blob_crates(blob: bytes) -> Generator[Crate, None, None]:
    d = {}
    for line in blob_lines(blob):
        crate = crate_from_text(line)
        d[crate.version] = crate
    for crate in d.values():
        yield crate


def crates_in_commit_range(
    start_commit: git.Commit, end_commit: git.Commit
) -> Generator[Crate, None, None]:
    """Generate newly specified crates."""

    rex = re.compile(
        r"""
        ^ 1/ [^/] $
        |
        ^ 2/ [^/]{2} $
        |
        ^ 3/ [^/] / [^/]{3} $
        |
        ^ [^/]{2} / [^/]{2} / [^/]{4,} $
        """,
        re.VERBOSE,
    )
    for path, start_blob, end_blob in blobs_in_commit_range(
        start_commit, end_commit
    ):
        if not rex.search(path):
            continue
        old_versions = set(crate.version for crate in blob_crates(start_blob))
        for crate in blob_crates(end_blob):
            if crate.version not in old_versions:
                yield crate


def crates_in_range(
    repo: git.Repo, start: str, end: str
) -> Generator[Crate, None, None]:
    start_commit = repo.commit(start) if start else None
    end_commit = repo.commit(end)
    yield from crates_in_commit_range(start_commit, end_commit)


def git_bundle_create(
    repo: git.Repo, bundle_file_name: str, start: str, end: str
) -> None:
    args = [
        "create",
        bundle_file_name,
    ]
    if start:
        args.append("^" + start)
    if end == "master":
        bundle_end = end
    else:
        repo.create_head("bundle/master", end, force=True)
        bundle_end = "bundle/master"
    args.append(bundle_end)
    repo.git.bundle(args)


def get_index_path(index: str) -> Path:
    path = Path(index)
    if not path.is_dir():
        raise error.UsageError(
            "{} is not a valid index directory".format(path)
        )
    return path


def get_repo(index: str) -> git.Repo:
    path = get_index_path(index)
    return git.Repo(str(path))


def get_crates_root_path(crates_root: str) -> Path:
    path = Path(crates_root)
    if not path.is_dir():
        raise error.UsageError(
            "{} is not a valid crates directory".format(path)
        )
    return path


def fetch_origin(repo: git.Repo) -> None:
    origin = git.remote.Remote(repo, "origin")
    origin.fetch(force=True)
    repo.create_head("origin_master", "remotes/origin/master", force=True)


def get_origin_url(repo: git.Repo) -> str:
    origin = git.remote.Remote(repo, "origin")
    urls = list(origin.urls)
    if len(urls) == 1:
        return urls[0]
    else:
        raise error.UsageError(
            "INDEX remote ``origin`` must have exactly one ``URL``"
        )


def mark(repo: git.Repo, end: str) -> None:
    for branch in ["mark", "master"]:
        if branch == repo.head.reference.name:
            common.eprint(
                "Will not move branch {} (it is current HEAD)".format(
                    repr(branch)
                )
            )
        else:
            common.iprint(
                "Move branch {} to point to {}".format(repr(branch), repr(end))
            )
            repo.create_head("refs/heads/{}".format(branch), end, force=True)


def list_crates(crates: List[Crate]) -> None:
    for crate in crates:
        common.iprint(crate.rel_path().name)


def _process_crates(
    downloader: romt.download.Downloader,
    dl_template: Optional[str],
    crates: List[Crate],
    crates_root: Path,
    good_paths_log_path: str,
    bad_paths_log_path: str,
    *,
    keep_going: bool,
    assume_ok: bool
) -> None:
    good_paths_file = common.open_optional(good_paths_log_path, "w")
    bad_paths_file = common.open_optional(bad_paths_log_path, "w")

    num_good_paths = 0
    num_bad_paths = 0

    limiter = downloader.new_limiter()

    async def _process_one(crate: Crate) -> None:
        nonlocal num_good_paths, num_bad_paths
        rel_path = crate.rel_path()
        path = crates_root / rel_path
        is_good = False
        try:
            if dl_template is None:
                downloader.verify_hash(path, crate.hash)
            else:
                url = dl_template.format(
                    crate=crate.name, version=crate.version
                )
                await downloader.adownload_verify_hash(
                    url, path, crate.hash, assume_ok=assume_ok
                )
            is_good = True
        except error.DownloadError as e:
            common.eprint(
                "Error: Download failure for {}: {}".format(
                    e.name, e.exception
                )
            )
        except error.MissingFileError as e:
            common.eprint("Error: Missing {}".format(e.name))
        except error.IntegrityError as e:
            common.eprint(str(e))

        if is_good:
            num_good_paths += 1
            common.log(good_paths_file, path)
        else:
            num_bad_paths += 1
            common.log(bad_paths_file, path)

        limiter.release_on_behalf_of(crate)

    async def _process_inner() -> None:
        async with trio.open_nursery() as nursery:
            for crate in crates:
                await limiter.acquire_on_behalf_of(crate)
                nursery.start_soon(
                    _process_one, crate,
                )

    downloader.run_job(_process_inner)

    common.iprint(
        "{} bad paths, {} good paths".format(num_bad_paths, num_good_paths)
    )

    common.close_optional(good_paths_file)
    common.close_optional(bad_paths_file)

    if num_bad_paths > 0 and not keep_going:
        raise error.AbortError()


def download_crates(
    downloader: romt.download.Downloader,
    dl_template: str,
    crates: List[Crate],
    crates_root: Path,
    good_paths_log_path: str,
    bad_paths_log_path: str,
    *,
    keep_going: bool,
    assume_ok: bool
) -> None:
    _process_crates(
        downloader,
        dl_template,
        crates,
        crates_root,
        good_paths_log_path,
        bad_paths_log_path,
        keep_going=keep_going,
        assume_ok=assume_ok,
    )


def verify_crates(
    downloader: romt.download.Downloader,
    crates: List[Crate],
    crates_root: Path,
    good_paths_log_path: str,
    bad_paths_log_path: str,
    *,
    keep_going: bool,
    assume_ok: bool
) -> None:
    _process_crates(
        downloader,
        None,
        crates,
        crates_root,
        good_paths_log_path,
        bad_paths_log_path,
        keep_going=keep_going,
        assume_ok=assume_ok,
    )


def pack(
    crates: List[Crate],
    crates_root: Path,
    bundle_path: Optional[Path],
    archive_path: Path,
    keep_going: bool,
) -> None:
    num_good_paths = 0
    num_bad_paths = 0

    with common.tar_context(archive_path, "w") as tar_f:
        if bundle_path is not None:
            packed_name = INDEX_BUNDLE_PACKED_NAME
            common.vprint("[pack] {}".format(packed_name))
            tar_f.add(str(bundle_path), packed_name)
        for rel_path in sorted(crate.rel_path() for crate in crates):
            path = crates_root / rel_path
            packed_name = "crates/" + rel_path.as_posix()
            try:
                common.vprint("[pack] {}".format(rel_path.name))
                tar_f.add(str(path), packed_name)
                num_good_paths += 1
            except FileNotFoundError:
                num_bad_paths += 1
                common.eprint("Error: Missing {}".format(rel_path))
                if not keep_going:
                    raise error.AbortError()

    common.iprint(
        "{} bad paths, {} good paths".format(num_bad_paths, num_good_paths)
    )


def unpack(
    repo: git.Repo,
    crates_root: Path,
    bundle_path: Path,
    archive_path: Path,
    keep_going: bool,
) -> None:
    num_crates = 0
    crates_prefix = "crates/"
    found_bundle = False

    with common.tar_context(archive_path, "r") as tar_f:
        for tar_info in tar_f:
            if tar_info.isdir():
                continue
            elif tar_info.name == INDEX_BUNDLE_PACKED_NAME:
                found_bundle = True
                tar_info.name = str(bundle_path)
                common.vprint("[unpack] {}".format(tar_info.name))
                tar_f.extract(tar_info)

            elif tar_info.name.startswith(crates_prefix):
                num_crates += 1
                tar_info.name = tar_info.name[len(crates_prefix) :]
                common.vprint(
                    "[unpack] {}".format(os.path.basename(tar_info.name))
                )
                tar_f.extract(tar_info, str(crates_root))

            else:
                common.eprint(
                    "Unexpected archive member {}".format(tar_info.name)
                )
                if not keep_going:
                    raise error.AbortError()

    if not found_bundle:
        common.eprint("Missing {} in archive".format(INDEX_BUNDLE_PACKED_NAME))
        if not keep_going:
            raise error.AbortError()

    common.iprint("{} extracted crates".format(num_crates))


def _config_json_path(repo: git.Repo) -> Path:
    working_tree = Path(repo.working_tree_dir)
    config_path = working_tree / "config.json"
    return config_path


def read_config_json(repo: git.Repo) -> Optional[bytes]:
    config_path = _config_json_path(repo)
    initial_config = None  # type: Optional[bytes]
    if config_path.is_file():
        initial_config = config_path.read_bytes()
    return initial_config


def update_config_json(repo: git.Repo, config: bytes) -> None:
    old_config = read_config_json(repo)
    if old_config is None or config != old_config:
        config_path = _config_json_path(repo)
        common.vprint("update-config: {}".format(config_path))
        config_path.write_bytes(config)
        repo.index.add(str(config_path))
        repo.index.commit("Apply config.json adjustments")


def _upgrade_to_working(repo: git.Repo) -> None:
    working = git.Reference(repo, "refs/heads/working")
    if repo.head.reference.name != "working" and not working.is_valid():
        # Time to upgrade working tree to use "working" branch.
        common.eprint("""Upgrade index to use "working" branch as HEAD""")
        if repo.head.reference.is_valid():
            # Create "working" branch based on current HEAD.
            common.iprint(
                """Checkout new "working" branch from current HEAD"""
            )
            repo.create_head("refs/heads/working", "HEAD")
        repo.head.set_reference(working)


def merge_origin_master(repo: git.Repo) -> None:
    _upgrade_to_working(repo)
    initial_config = read_config_json(repo)

    try:
        common.vprint("merge-index: merge origin/master")
        repo.git.merge("remotes/origin/master", "-m", "Merge origin/master")
    except git.GitError:
        common.iprint("merge-index: merge failed; reconstructing")
        common.vprint("merge-index: reset to recover failed merge state")
        repo.head.reset(working_tree=True, index=True)
        common.vprint("merge-index: reset to remotes/origin/master")
        repo.head.reset("remotes/origin/master", working_tree=True, index=True)

    # Restore initial_config if necessary.
    if initial_config is not None:
        update_config_json(repo, initial_config)


def _init_common(
    index_path: Path, origin_location: str, crates_root_path: Path
) -> git.Repo:
    if index_path.is_dir():
        raise error.UsageError(
            "index directory {} already exists".format(repr(str(index_path)))
        )
    common.iprint(
        "create index repository at {}:".format(repr(str(index_path)))
    )
    index_path.mkdir(parents=True)
    repo = git.Repo.init(str(index_path))
    common.iprint("  remote add origin {}".format(origin_location))
    repo.create_remote("origin", origin_location)

    # Setup "HEAD" to new "working" branch.
    working = git.Reference(repo, "refs/heads/working")
    repo.head.set_reference(working)

    # Setup default remote and merge branch for "working".
    with repo.config_writer() as writer:
        writer.set_value('branch "working"', "remote", "origin")
        writer.set_value('branch "working"', "merge", "refs/heads/master")

    if not crates_root_path.is_dir():
        common.iprint(
            "create crates directory at {}:".format(
                repr(str(crates_root_path))
            )
        )
        crates_root_path.mkdir(parents=True)
    return repo


def init(index_path: Path, index_url: str, crates_root_path: Path) -> None:
    _init_common(index_path, index_url, crates_root_path)


def init_import(index_path: Path, crates_root_path: Path) -> None:
    bundle_path = index_path / INDEX_BUNDLE_NAME
    bundle_location = str(bundle_path.absolute())
    repo = _init_common(index_path, bundle_location, crates_root_path)
    with repo.config_writer() as writer:
        writer.set_value(
            'remote "origin"', "fetch", "+refs/heads/*:refs/remotes/origin/*"
        )
        writer.add_value(
            'remote "origin"',
            "fetch",
            "+refs/heads/bundle/*:refs/remotes/origin/*",
        )


def configure_index(repo: git.Repo, server_url: str) -> None:
    if not server_url.endswith("/"):
        server_url += "/"
    config = """\
{{
    "dl": "{}",
    "api": "{}"
}}
""".format(
        server_url + "crates/{crate}/{crate}-{version}.crate", server_url
    )
    update_config_json(repo, config.encode("utf-8"))


def add_arguments(parser: argparse.ArgumentParser) -> None:
    base.add_downloader_arguments(parser)

    parser.add_argument(
        "--index",
        action="store",
        default=str(INDEX_STANDARD_PATH),
        help="""{INDEX_NAME} repository path (default: %(default)s)""".format(
            INDEX_NAME=INDEX_NAME
        ),
    )

    parser.add_argument(
        "--crates",
        action="store",
        metavar="CRATES_ROOT",
        default="crates",
        help="directory holding the crates (default: %(default)s)",
    )

    parser.add_argument(
        "--start",
        action="store",
        help="""reference to start of RANGE (``0`` for start of repo);
            if not provided START defaults to ``mark`` and
            --allow-missing-start is implied""",
    )

    parser.add_argument(
        "--end",
        action="store",
        default="HEAD",
        help="reference to end of RANGE (default: %(default)s)",
    )

    parser.add_argument(
        "--allow-missing-start",
        action="store_true",
        help="treat non-existent START as start of repo instead of an error",
    )

    parser.add_argument(
        "--good-paths",
        action="store",
        metavar="GOOD_PATHS",
        help="record successful paths to file GOOD_PATHS",
    )

    parser.add_argument(
        "--bad-paths",
        action="store",
        metavar="BAD_PATHS",
        help="record bad paths to file BAD_PATHS",
    )

    parser.add_argument(
        "--archive",
        action="store",
        default="crates.tar.gz",
        help="use file ARCHIVE for pack/unpack (default: %(default)s)",
    )

    parser.add_argument(
        "--keep-going",
        action="store_true",
        default=False,
        help="keep going even if errors occur (helps with missing crates)",
    )

    parser.add_argument(
        "--crates-url",
        action="store",
        default=(
            "https://static.crates.io/crates/{crate}/{crate}-{version}.crate"
        ),
        help="""template for crates download URL for ``download`` command;
            use {crate} and {version} to parametrize with crate's name
            and version number; to use the API as defined in
            crates.io-index/config.json, use
            "https://crates.io/api/v1/crates/{crate}/{version}/download";
            default=%(default)s
        """,
    )

    parser.add_argument(
        "--index-url",
        action="store",
        default="https://github.com/rust-lang/crates.io-index",
        help="""URL of upstream crates.io-index Git repository
            for ``init`` (default: %(default)s)""",
    )

    parser.add_argument(
        "--bundle-path",
        action="store",
        help="""local path to store {INDEX_BUNDLE_NAME} for ``pack``,
            ``unpack``; configures path to origin repo for ``init-import``;
            (default: INDEX/{INDEX_BUNDLE_NAME})""".format(
            INDEX_BUNDLE_NAME=INDEX_BUNDLE_NAME
        ),
    )

    parser.add_argument(
        "--server-url",
        action="store",
        default=("http://localhost:8000"),
        help="""base URL for server configured into INDEX/config.json by
            ``config`` command (default=%(default)s)
        """,
    )

    parser.add_argument(
        "commands",
        nargs="*",
        metavar="COMMAND",
        help="commands to execute in the order given",
    )


class Main(base.BaseMain):
    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self._repo = None  # type: Optional[git.Repo]
        self._crates = None  # type: Optional[List[Crate]]

    def _get_start(self) -> str:
        start = self.args.start
        if start is None:
            raise error.UsageError("missing START")
        elif start == "0":
            start = ""
        return start

    def get_start(self) -> str:
        start = self._get_start()
        if (
            start
            and self.args.allow_missing_start
            and not self.branch_exists(start)
        ):
            start = ""
        return start

    def get_start_branch_name(self) -> str:
        start = self._get_start()
        if start == "":
            raise error.UsageError("START may not be 0 or empty")
        return start

    def get_repo(self) -> git.Repo:
        if self._repo is None:
            self._repo = get_repo(self.args.index)
        return self._repo

    def branch_exists(self, branch: str) -> bool:
        try:
            self.get_repo().commit(branch)
        except git.BadName:
            return False
        return True

    def forget_crates(self) -> None:
        self._crates = None

    def get_crates(self) -> List[Crate]:
        if self._crates is None:
            common.vprint("[calculating crate list]")
            self._crates = list(
                crates_in_range(
                    self.get_repo(), self.get_start(), self.args.end
                )
            )
            common.vprint("[{} crates in range]".format(len(self._crates)))
        return self._crates

    def get_bundle_path(self) -> Path:
        if self.args.bundle_path:
            path = Path(self.args.bundle_path)
        elif self.args.index:
            path = get_index_path(self.args.index) / INDEX_BUNDLE_NAME
        else:
            raise error.UsageError("missing BUNDLE_PATH")
        # Use .absolute() because git operations change the working directory.
        return path.absolute()

    def get_origin_bundle_path(self) -> Path:
        """Path for repo's remote "origin" URL"""
        url = get_origin_url(self.get_repo())
        parsed = urllib.parse.urlparse(url)
        if parsed.scheme or parsed.netloc:
            raise error.UsageError(
                "INDEX remote ``origin`` must have ``url`` as a local file"
            )
        else:
            path = Path(url)

        if self.args.bundle_path and (
            path.resolve() != Path(self.args.bundle_path).resolve()
        ):
            raise error.UsageError(
                "BUNDLE_PATH must match ``url`` for INDEX's ``origin`` remote"
            )

        return path

    def get_commands(self) -> List[str]:
        valid_commands = [
            "pull",
            "download",
            "verify",
            "pack",
            "mark",
            "unpack",
            "list",
            "update",
            "export",
            "import",
            "init",
            "init-import",
            "config",
        ]

        commands = self.args.commands[:]
        if commands:
            base.verify_commands(commands, valid_commands)
        else:
            common.iprint("Nothing to do (try --help)")
        return commands

    def cmd_pull(self) -> None:
        self.forget_crates()
        repo = self.get_repo()
        fetch_origin(repo)
        merge_origin_master(repo)

    def cmd_mark(self) -> None:
        self.forget_crates()
        mark(
            self.get_repo(), self.args.end,
        )

    def cmd_unpack(self) -> None:
        self.forget_crates()
        unpack(
            self.get_repo(),
            get_crates_root_path(self.args.crates),
            self.get_origin_bundle_path(),
            self.get_archive_path(),
            self.args.keep_going,
        )

    def cmd_list(self) -> None:
        list_crates(self.get_crates())

    def cmd_download(self) -> None:
        download_crates(
            self.downloader,
            self.args.crates_url,
            self.get_crates(),
            get_crates_root_path(self.args.crates),
            self.args.good_paths,
            self.args.bad_paths,
            keep_going=self.args.keep_going,
            assume_ok=self.args.assume_ok,
        )

    def cmd_verify(self) -> None:
        verify_crates(
            self.downloader,
            self.get_crates(),
            get_crates_root_path(self.args.crates),
            self.args.good_paths,
            self.args.bad_paths,
            keep_going=self.args.keep_going,
            assume_ok=self.args.assume_ok,
        )

    def cmd_pack(self) -> None:
        if self.get_crates():
            bundle_path = self.get_bundle_path()
            git_bundle_create(
                self.get_repo(),
                str(bundle_path),
                self.get_start(),
                self.args.end,
            )

            pack(
                self.get_crates(),
                get_crates_root_path(self.args.crates),
                bundle_path,
                self.get_archive_path(),
                self.args.keep_going,
            )
        else:
            common.iprint("No crates to pack")

    def cmd_init(self) -> None:
        index_path = Path(self.args.index)
        index_url = self.args.index_url
        parsed = urllib.parse.urlparse(index_url)
        if not parsed.scheme and not parsed.netloc:
            # This is not a URL, but a local path.
            # Relative local directories in Git remotes can cause trouble,
            # because when they are stored in the .git/config file they are
            # interpreted to be relative to the .git/ directory, which changes
            # their meaning.
            index_url = os.path.abspath(index_url)
        init(index_path, index_url, Path(self.args.crates))

    def cmd_init_import(self) -> None:
        index_path = Path(self.args.index)
        init_import(index_path, Path(self.args.crates))

    def cmd_config(self) -> None:
        self.forget_crates()
        configure_index(self.get_repo(), self.args.server_url)
        mark(self.get_repo(), self.args.end)

    def _run(self) -> None:
        if not self.args.start:
            self.args.start = "mark"
            self.args.allow_missing_start = True

        commands = self.get_commands()

        while commands:
            command = commands.pop(0)
            common.iprint("{}...".format(command))
            if command in ("update", "export", "import"):
                if command == "update":
                    cmd = "pull download mark"
                elif command == "export":
                    cmd = "pull download pack mark"
                else:
                    cmd = "unpack pull verify mark"
                commands = cmd.split() + commands

            else:
                cmd_func_name = "cmd_{}".format(command.replace("-", "_"))
                getattr(self, cmd_func_name)()



#!/usr/bin/env python3
# coding=utf-8

from typing import Optional


class Error(Exception):
    def __init__(self, message: str) -> None:
        super().__init__()
        self.message = message

    def __str__(self) -> str:
        return "error: {}".format(self.message)


class UsageError(Error):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class AbortError(Error):
    def __init__(self, message: str = "") -> None:
        if message:
            super().__init__("Aborting: {}".format(message))
        else:
            super().__init__("Aborting")


class IntegrityError(Error):
    def __init__(
        self,
        name: str,
        *,
        actual_hash: Optional[str] = None,
        expected_hash: Optional[str] = None,
        sig_name: Optional[str] = None
    ) -> None:
        if actual_hash and expected_hash:
            message = "Bad hash for {} (got {}, expected {})".format(
                name, actual_hash, expected_hash,
            )
        elif sig_name:
            message = "Signature failure for {} using signature {}".format(
                name, sig_name
            )
        else:
            message = "Integrity failure"
        super().__init__(message)
        self.name = name
        self.actual_hash = actual_hash
        self.expected_hash = expected_hash
        self.sig_name = sig_name


class MissingFileError(Error):
    def __init__(self, name: str) -> None:
        super().__init__("missing file {}".format(name))
        self.name = name


class MissingDirectoryError(Error):
    def __init__(self, name: str) -> None:
        super().__init__("missing directory {}".format(name))
        self.name = name


class DownloadError(Error):
    def __init__(self, name: str, exception: Exception) -> None:
        super().__init__("failed to download {}".format(name))
        self.name = name
        self.exception = exception


class UnexpectedArchiveMemberError(Error):
    def __init__(self, name: str) -> None:
        super().__init__("unexpected archive member {}".format(name))
        self.name = name



#!/usr/bin/env python3
# coding=utf-8

import functools
from pathlib import Path
from typing import (
    Any,
    Awaitable,
    BinaryIO,
    Callable,
)

import httpx
import trio

from romt import common
from romt import error
from romt import integrity
from romt import signature


class Downloader:
    def __init__(self, num_jobs: int) -> None:
        self._client = httpx.AsyncClient()
        self.sig_verifier = signature.Verifier()
        self._warn_signature = False
        self.num_jobs = num_jobs

    def set_warn_signature(self, warn_signature: bool) -> None:
        self._warn_signature = warn_signature

    async def adownload_fileobj(self, url: str, fileobj: BinaryIO) -> None:
        try:
            async with self._client.stream("GET", url) as response:
                response.raise_for_status()
                async for chunk in response.aiter_bytes(chunk_size=16384):
                    fileobj.write(chunk)
        except httpx.HTTPError as e:
            raise error.DownloadError(url, e)

    def download_fileobj(self, url: str, fileobj: BinaryIO) -> None:
        self.run_job(self.adownload_fileobj, url, fileobj)

    async def _adownload(self, url: str, dest_path: Path) -> None:
        common.make_dirs_for(dest_path)
        try:
            with dest_path.open("wb") as f:
                await self.adownload_fileobj(url, f)
        except error.DownloadError:
            if dest_path.is_file():
                dest_path.unlink()
            raise

    async def adownload(self, url: str, dest_path: Path) -> None:
        if dest_path.is_file():
            dest_path.unlink()
        tmp_dest_path = common.tmp_path_for(dest_path)
        await self._adownload(url, tmp_dest_path)
        tmp_dest_path.rename(dest_path)

    def download(self, url: str, dest_path: Path) -> None:
        self.run_job(self.adownload, url, dest_path)

    async def adownload_cached(
        self, dest_url: str, dest_path: Path, *, cached: bool = True
    ) -> None:
        if cached and dest_path.is_file():
            common.vprint("[cached file] {}".format(dest_path))
        else:
            common.vprint("[downloading] {}".format(dest_path))
            await self.adownload(dest_url, dest_path)

    def download_cached(
        self, dest_url: str, dest_path: Path, *, cached: bool = True
    ) -> None:
        self.run_job(self.adownload_cached, dest_url, dest_path, cached=cached)

    def _sig_verify(self, path: Path, sig_path: Path) -> None:
        try:
            self.sig_verifier.verify(path, sig_path)
        except (error.MissingFileError, error.IntegrityError):
            if self._warn_signature:
                common.eprint("Warning: Signature failure for {}".format(path))
            else:
                raise

    def verify_hash(self, path: Path, hash: str) -> None:
        """
        Raises:
            MissingFileError - path doesn't exist
            IntegrityError - path exists with bad hash
        """
        common.vprint("[verify] {}".format(path))
        integrity.verify_hash(path, hash)

    def verify(self, path: Path, *, with_sig: bool = False) -> None:
        """
        Raises:
            MissingFileError - path or associated sha256path doesn't exist
            IntegrityError - paths exists with bad hash
        """
        hash_path = integrity.path_append_hash_suffix(path)
        common.vprint("[verify] {}".format(path))
        integrity.verify(path, hash_path)
        if with_sig:
            sig_path = signature.path_append_sig_suffix(path)
            self._sig_verify(path, sig_path)

    async def adownload_verify_hash(
        self,
        dest_url: str,
        dest_path: Path,
        hash: str,
        *,
        cached: bool = True,
        assume_ok: bool = False
    ) -> None:
        if cached:
            if assume_ok and dest_path.is_file():
                common.vvprint("[assuming OK] {}".format(dest_path))
                return
            try:
                integrity.verify_hash(dest_path, hash)
                common.vprint("[cached file] {}".format(dest_path))
                return
            except (error.MissingFileError, error.IntegrityError):
                pass
        common.vprint("[downloading] {}".format(dest_path))
        await self.adownload(dest_url, dest_path)
        integrity.verify_hash(dest_path, hash)

    def download_verify_hash(
        self,
        dest_url: str,
        dest_path: Path,
        hash: str,
        *,
        cached: bool = True,
        assume_ok: bool = False
    ) -> None:
        self.run_job(
            self.adownload_verify_hash,
            dest_url,
            dest_path,
            hash,
            cached=cached,
            assume_ok=assume_ok,
        )

    async def adownload_verify(
        self,
        dest_url: str,
        dest_path: Path,
        *,
        cached: bool = True,
        assume_ok: bool = False,
        with_sig: bool = False
    ) -> None:
        hash_path = integrity.path_append_hash_suffix(dest_path)
        sig_path = signature.path_append_sig_suffix(dest_path)
        if cached:
            if (
                assume_ok
                and dest_path.is_file()
                and hash_path.is_file()
                and (not with_sig or sig_path.is_file())
            ):
                common.vvprint("[assuming OK] {}".format(dest_path))
                return
            try:
                integrity.verify(dest_path, hash_path)
                if with_sig:
                    self._sig_verify(dest_path, sig_path)
                common.vprint("[cached file] {}".format(dest_path))
                return
            except (error.MissingFileError, error.IntegrityError):
                pass
        common.vprint("[downloading] {}".format(dest_path))
        # Download the (small) hash and signature files first.
        hash_url = integrity.append_hash_suffix(dest_url)
        await self.adownload(hash_url, hash_path)
        if with_sig:
            sig_url = signature.append_sig_suffix(dest_url)
            await self.adownload(sig_url, sig_path)

        # If dest_path exists and has the correct hash, bypass the downloading
        # step to save download time.
        download_required = True
        if dest_path.is_file():
            try:
                integrity.verify(dest_path, hash_path)
                download_required = False
            except (error.MissingFileError, error.IntegrityError):
                pass
        if download_required:
            await self.adownload(dest_url, dest_path)
            integrity.verify(dest_path, hash_path)

        if with_sig:
            self._sig_verify(dest_path, sig_path)

    def download_verify(
        self,
        dest_url: str,
        dest_path: Path,
        *,
        cached: bool = True,
        assume_ok: bool = False,
        with_sig: bool = False
    ) -> None:
        self.run_job(
            self.adownload_verify,
            dest_url,
            dest_path,
            cached=cached,
            assume_ok=assume_ok,
            with_sig=with_sig,
        )

    def run_job(
        self, job: Callable[..., Awaitable[None]], *args: Any, **kwargs: Any
    ) -> None:
        trio.run(functools.partial(job, **kwargs), *args)

    def new_limiter(self) -> trio.CapacityLimiter:
        return trio.CapacityLimiter(self.num_jobs)

    def close(self) -> None:
        self.run_job(self._client.aclose)



#!/usr/bin/env python3
# coding=utf-8

import argparse

from pathlib import Path
from typing import List, Optional
import urllib.parse

from romt import base
from romt import common
from romt import error


def expand_target_alias(target: str) -> str:
    if target == "linux":
        target = "x86_64-unknown-linux-gnu"
    elif target == "windows":
        target = "x86_64-pc-windows-msvc"
    elif target == "darwin":
        target = "x86_64-apple-darwin"
    return target


def target_exe_suffix(target: str) -> str:
    parts = target.split("-")
    if "windows" in parts:
        suffix = ".exe"
    else:
        suffix = ""
    return suffix


def append_exe_suffix(s: str, target: str) -> str:
    suffix = target_exe_suffix(target)
    return s + suffix


def path_append_exe_suffix(path: Path, target: str) -> Path:
    suffix = target_exe_suffix(target)
    return common.path_append(path, suffix)


def require_specs(specs: List[str]) -> List[str]:
    if not specs:
        raise error.UsageError("missing required SPEC; try --select")
    return specs


def require_targets(
    targets: List[str], *, default: Optional[str] = None
) -> List[str]:
    if not targets:
        if default is None:
            raise error.UsageError("missing required TARGET; try --target")
        targets = [default]
    return targets


class DistMain(base.BaseMain):
    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self._specs = None  # type: Optional[List[str]]
        self._targets = None  # type: Optional[List[str]]

    @property
    def specs(self) -> List[str]:
        if self._specs is None:
            specs = common.split_flatten_words(self.args.specs)
            self._specs = specs
        return self._specs

    @specs.setter
    def specs(self, value: List[str]) -> None:
        self._specs = list(value)

    @property
    def targets(self) -> List[str]:
        if self._targets is None:
            patterns = [
                expand_target_alias(pattern)
                for pattern in common.normalize_patterns(self.args.targets)
            ]
            self._targets = patterns
        return self._targets

    @targets.setter
    def targets(self, targets: List[str]) -> None:
        self._targets = list(targets)

    @property
    def dest_path(self) -> Path:
        return Path(self.args.dest)

    def dest_path_from_rel_path(self, rel_path: str) -> Path:
        return self.dest_path / rel_path

    def url_from_rel_path(self, rel_path: str) -> str:
        base_url = self.args.url
        if not base_url.endswith("/"):
            base_url += "/"
        return urllib.parse.urljoin(base_url, rel_path)
